#include <iostream>
#include <chrono>
#include <vector>
#include <utility>
#include <thread>
#include <tuple>
#include <functional>

using namespace std; 

void dummy(int i) {
  cout << "Thread no. " << i << endl;
  return;
}

void mm(int istart, int iend, int ncol,
	vector<vector<float>>&a,
	vector<vector<float>>&b,
	vector<vector<float>>&c
	) {
  for(int i=istart; i<iend; i++)
    for(int j=0; j<ncol; j++) {
      c[i][j] =0.0;
      for(int k=0; k<ncol; k++)
	c[i][j] += a[i][k]*b[k][j];
    }
  return; 
}

void print(vector<vector<float>>& a, int n) {
  for(int i=0; i<n; i++) {
    for(int j=0; j<n; j++) {
      cout << " " << a[i][j];
    }
    cout << endl;
  }
  return;
}

void init_rand(vector<vector<float>>& a, int n, int seed, int max) {
  srand(seed);        
  for(int i=0; i<n; i++) 
    for(int j=0; j<n; j++) {
      a[i][j] = ((float) (rand() % max));
    }
  return;
}

void init_const(vector<vector<float>>& a, int n, float x) {
  for(int i=0; i<n; i++)      
    for(int j=0; j<n; j++) {
      a[i][j] = x;
    }
  return;
}


int main(int argc, char * argv[]) {
  if(argc == 1) {
    cout << "Usage is: " << argv[0] << "  N nw " <<endl;
    return(-1);
  }
  int n = atoi(argv[1]);   // dimension of the matrix
  int t = atoi(argv[2]);   // number of threads
  int seed = 123;
  int max  = 16;

  cout << "Initializing 4 " << n << "x" << n << " matrixes " << endl; 
  vector<vector<float>> a(n,vector<float>(n));
  init_rand(a, n, seed, max);
  // print(a,n);
  
  vector<vector<float>> b(n,vector<float>(n));
  init_rand(b, n, seed, max);
  // print(b,n);
  
  vector<vector<float>> c(n,vector<float>(n));
  init_const(c,n,0.0);

  vector<vector<float>> d(n,vector<float>(n));
  init_const(d,n,0.0);


  cout << "Start sequential computation ..." << endl;
  
  auto start = std::chrono::system_clock::now();
  for(int i=0; i<n; i++)
    for(int j=0; j<n; j++)
      for(int k=0; k<n; k++)
  	d[i][j] += a[i][k]*b[k][j];
  auto end = std::chrono::system_clock::now();
  chrono::duration<double> tseq = end-start;

  cout << "Ended" << endl;
  cout << "Start openmp parallel computation ..." << endl; 

  start = std::chrono::system_clock::now();
#pragma omp parallel for num_threads(t)
   for(int i=0; i<n; i++)
     for(int j=0; j<n; j++)
       for(int k=0; k<n; k++)
   	c[i][j] += a[i][k]*b[k][j];
  end = std::chrono::system_clock::now();
  chrono::duration<double> tomp = end-start;
  cout << "Ended. Checking ... " << endl; 
  bool same = true; 
  for(int i=0; i<n; i++)
    for(int j=0; j<n; j++)
      if(c[i][j] != d[i][j])
	same = false;
  if(!same)
    cout << "par openmp and seq result differ" << endl;
  else
    cout << "Passed!" << endl;
  
  cout << "Starting parallel thread computation ...  " << endl; 
  // now fork threads computing the ranks
  vector<thread> tid;
  cout << "Computing ranges ..." << endl; 
  start = std::chrono::system_clock::now();
  // prepare ranks for parallel computation
  vector<pair<int,int>> ranges;
  int k = n / t;
  int r = n % t;
  for(int i=0; i<t; i++) {
    if(i==(t-1))
      ranges.push_back(pair<int,int>(i*k, (i+1)*k+r));
    else
      ranges.push_back(pair<int,int>(i*k, (i+1)*k));
  }
  // for(int i=0; i<t; i++) {
  //   cout << ranges[i].first << "," << ranges[i].second << " ";
  // }
  // cout << endl;

  // fork threads
  cout << "Forking threads ..." << endl; 
  for(int i=0; i<t; i++)  {
    tid.push_back(thread(mm,              // the function
			 ranges[i].first, // start row
			 ranges[i].second,// end row
			 n,               // len of row/col
			 ref(a), ref(b), ref(c) // the matrixes
			 ));        
  }
  // await termination (kind of barrier...)
  cout << "Joining threads ... " << endl; 
  for(int i=0; i<t; i++)  {
    tid[i].join();
  }

  cout << "Done ! " << endl;
  
  end = std::chrono::system_clock::now();
  chrono::duration<double> tpar = end-start;
  same = true; 
  for(int i=0; i<n; i++)
    for(int j=0; j<n; j++)
      if(c[i][j] != d[i][j])
	same = false;
  if(!same)
    cout << "par thread and seq result differ" << endl;


  auto perfs = [&](chrono::duration<double> time, int n, string typ) {
    double flops = ((double) (n*n*n)) / time.count();
    cout << "mm " << typ << " took " << time.count() << " seconds delivering " <<
    (flops/ ((double) (1024*1024*1024)) )<< " GFlops with " << t << " threads " << endl;
  };

  perfs(tseq,n,"seq");
  perfs(tomp,n,"omp");
  perfs(tpar,n,"par");
  
  cout << "Computed with ranges : " << endl; 
  for(int i=0; i<t; i++) {
    cout << ranges[i].first << "," << ranges[i].second << " (" <<
      ranges[i].second - ranges[i].first << ")";
  }
  cout << endl;

  return(0);
  
}

